﻿using System;
using HeuristicLab.Algorithms.OffspringSelectionGeneticAlgorithm;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.Xml;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using HeuristicLab.Selection;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Linq;
using HeuristicLab.Problems.Instances.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Optimization;
using HeuristicLab.ParallelEngine;

namespace UnitTests {
  [TestClass]
  public class PerformanceTest {
    private static int seed = 1234;

    //TODO remove performance test 
    [TestMethod]
    [TestCategory("Problems.DataAnalysis.Symbolic.Regression")]
    [TestProperty("Time", "long")]
    public void ConstantsOptimization_Tower_Algorithm() {
      var algorithm = CreateGpSymbolicRegressionSample();
      var problem = (SymbolicRegressionSingleObjectiveProblem)algorithm.Problem;

      algorithm.MaximumGenerations.Value = 1;
      var evaluator = new ConstantsOptimizationEvaluator();
      evaluator.ConstantOptimizationIterations.Value = 5;
      problem.EvaluatorParameter.Value = evaluator;

      algorithm.Prepare();
      algorithm.Start();
    }

    private OffspringSelectionGeneticAlgorithm CreateGpSymbolicRegressionSample() {
      var osga = new OffspringSelectionGeneticAlgorithm();
      #region Problem Configuration
      var provider = new RegressionRealWorldInstanceProvider();
      var problemData =  provider.LoadData(new Tower());
      var problem = new SymbolicRegressionSingleObjectiveProblem();
      problem.ProblemData = problemData;
      problem.Load(problemData);
      problem.BestKnownQuality.Value = 1.0;

      #region configure grammar

      var grammar = (TypeCoherentExpressionGrammar)problem.SymbolicExpressionTreeGrammar;
      grammar.ConfigureAsDefaultRegressionGrammar();

      //symbols square, power, squareroot, root, log, exp, sine, cosine, tangent, variable
      var square = grammar.Symbols.OfType<Square>().Single();
      var power = grammar.Symbols.OfType<Power>().Single();
      var squareroot = grammar.Symbols.OfType<SquareRoot>().Single();
      var root = grammar.Symbols.OfType<Root>().Single();
      var cube = grammar.Symbols.OfType<Cube>().Single();
      var cuberoot = grammar.Symbols.OfType<CubeRoot>().Single();
      var log = grammar.Symbols.OfType<Logarithm>().Single();
      var exp = grammar.Symbols.OfType<Exponential>().Single();
      var sine = grammar.Symbols.OfType<Sine>().Single();
      var cosine = grammar.Symbols.OfType<Cosine>().Single();
      var tangent = grammar.Symbols.OfType<Tangent>().Single();
      var variable = grammar.Symbols.OfType<HeuristicLab.Problems.DataAnalysis.Symbolic.Variable>().Single();
      var powerSymbols = grammar.Symbols.Single(s => s.Name == "Power Functions");
      powerSymbols.Enabled = true;

      square.Enabled = true;
      square.InitialFrequency = 1.0;
      foreach (var allowed in grammar.GetAllowedChildSymbols(square))
        grammar.RemoveAllowedChildSymbol(square, allowed);
      foreach (var allowed in grammar.GetAllowedChildSymbols(square, 0))
        grammar.RemoveAllowedChildSymbol(square, allowed, 0);
      grammar.AddAllowedChildSymbol(square, variable);

      power.Enabled = false;

      squareroot.Enabled = false;
      foreach (var allowed in grammar.GetAllowedChildSymbols(squareroot))
        grammar.RemoveAllowedChildSymbol(squareroot, allowed);
      foreach (var allowed in grammar.GetAllowedChildSymbols(squareroot, 0))
        grammar.RemoveAllowedChildSymbol(squareroot, allowed, 0);
      grammar.AddAllowedChildSymbol(squareroot, variable);

      cube.Enabled = false;
      cuberoot.Enabled = false;
      root.Enabled = false;

      log.Enabled = true;
      log.InitialFrequency = 1.0;
      foreach (var allowed in grammar.GetAllowedChildSymbols(log))
        grammar.RemoveAllowedChildSymbol(log, allowed);
      foreach (var allowed in grammar.GetAllowedChildSymbols(log, 0))
        grammar.RemoveAllowedChildSymbol(log, allowed, 0);
      grammar.AddAllowedChildSymbol(log, variable);

      exp.Enabled = true;
      exp.InitialFrequency = 1.0;
      foreach (var allowed in grammar.GetAllowedChildSymbols(exp))
        grammar.RemoveAllowedChildSymbol(exp, allowed);
      foreach (var allowed in grammar.GetAllowedChildSymbols(exp, 0))
        grammar.RemoveAllowedChildSymbol(exp, allowed, 0);
      grammar.AddAllowedChildSymbol(exp, variable);

      sine.Enabled = false;
      foreach (var allowed in grammar.GetAllowedChildSymbols(sine))
        grammar.RemoveAllowedChildSymbol(sine, allowed);
      foreach (var allowed in grammar.GetAllowedChildSymbols(sine, 0))
        grammar.RemoveAllowedChildSymbol(sine, allowed, 0);
      grammar.AddAllowedChildSymbol(sine, variable);

      cosine.Enabled = false;
      foreach (var allowed in grammar.GetAllowedChildSymbols(cosine))
        grammar.RemoveAllowedChildSymbol(cosine, allowed);
      foreach (var allowed in grammar.GetAllowedChildSymbols(cosine, 0))
        grammar.RemoveAllowedChildSymbol(cosine, allowed, 0);
      grammar.AddAllowedChildSymbol(cosine, variable);

      tangent.Enabled = false;
      foreach (var allowed in grammar.GetAllowedChildSymbols(tangent))
        grammar.RemoveAllowedChildSymbol(tangent, allowed);
      foreach (var allowed in grammar.GetAllowedChildSymbols(tangent, 0))
        grammar.RemoveAllowedChildSymbol(tangent, allowed, 0);
      grammar.AddAllowedChildSymbol(tangent, variable);
      #endregion

      problem.SymbolicExpressionTreeGrammar = grammar;

      // configure remaining problem parameters
      problem.MaximumSymbolicExpressionTreeLength.Value = 50;
      problem.MaximumSymbolicExpressionTreeDepth.Value = 12;
      problem.MaximumFunctionDefinitions.Value = 0;
      problem.MaximumFunctionArguments.Value = 0;

      var evaluator = new SymbolicRegressionConstantOptimizationEvaluator();
      evaluator.ConstantOptimizationIterations.Value = 5;
      problem.EvaluatorParameter.Value = evaluator;
      problem.RelativeNumberOfEvaluatedSamplesParameter.Hidden = true;
      problem.SolutionCreatorParameter.Hidden = true;
      #endregion

      #region Algorithm Configuration
      osga.Problem = problem;
      osga.Name = "Offspring Selection Genetic Programming - Symbolic Regression";
      osga.Description = "Genetic programming with strict offspring selection for solving a benchmark regression problem.";
      ConfigureOsGeneticAlgorithmParameters<GenderSpecificSelector, SubtreeCrossover, MultiSymbolicExpressionTreeManipulator>(osga, 100, 1, 25, 0.2, 50);
      var mutator = (MultiSymbolicExpressionTreeManipulator)osga.Mutator;
      mutator.Operators.OfType<FullTreeShaker>().Single().ShakingFactor = 0.1;
      mutator.Operators.OfType<OnePointShaker>().Single().ShakingFactor = 1.0;

      osga.Analyzer.Operators.SetItemCheckedState(
        osga.Analyzer.Operators
          .OfType<SymbolicRegressionSingleObjectiveOverfittingAnalyzer>()
          .Single(), false);
      osga.Analyzer.Operators.SetItemCheckedState(
        osga.Analyzer.Operators
          .OfType<SymbolicDataAnalysisAlleleFrequencyAnalyzer>()
          .First(), false);

      osga.ComparisonFactorModifierParameter.Hidden = true;
      osga.ComparisonFactorLowerBoundParameter.Hidden = true;
      osga.ComparisonFactorUpperBoundParameter.Hidden = true;
      osga.OffspringSelectionBeforeMutationParameter.Hidden = true;
      osga.SuccessRatioParameter.Hidden = true;
      osga.SelectedParentsParameter.Hidden = true;
      osga.ElitesParameter.Hidden = true;

      #endregion
      return osga;
    }


    public static void ConfigureOsGeneticAlgorithmParameters<S, C, M>(OffspringSelectionGeneticAlgorithm ga, int popSize, int elites, int maxGens, double mutationRate = 0.05, double maxSelPres = 100, int tournGroupSize = 0)
      where S : ISelector
      where C : ICrossover
      where M : IManipulator {
      ga.Elites.Value = elites;
      ga.MaximumGenerations.Value = maxGens;
      ga.MutationProbability.Value = mutationRate;
      ga.PopulationSize.Value = popSize;
      ga.MaximumSelectionPressure.Value = maxSelPres;
      ga.Seed.Value = 0;
      ga.SetSeedRandomly.Value = true;
      ga.ComparisonFactorLowerBound.Value = 1;
      ga.ComparisonFactorUpperBound.Value = 1;

      ga.Selector = ga.SelectorParameter.ValidValues
        .OfType<S>()
        .First();

      ga.Crossover = ga.CrossoverParameter.ValidValues
        .OfType<C>()
        .First();

      ga.Mutator = ga.MutatorParameter.ValidValues
        .OfType<M>()
        .First();

      var tSelector = ga.Selector as TournamentSelector;
      if (tSelector != null) {
        tSelector.GroupSizeParameter.Value.Value = tournGroupSize;
      }
      ga.Engine = new ParallelEngine();
    }

  }
}
